/*
Copyright (c) 2015, Movania Muhammad Mobeen
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list
of conditions and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice, this list
of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

//Demo source code for Game Engine Gems 3 chapter "Simulating Soft Bodies using Strain
//Based Dynamics" by Dr. Muhammad Mobeen Movania. For details about the method,
//please refer to the book chapter. This project implements cloth using strain based 
//dynamics based on the paper by Mueller et. al.
//http://matthias-mueller-fischer.ch/publications/strainBasedDynamics.pdf using 
//GLUT,GLEW and GLM libraries. This code is intended for beginners so that they may 
//understand what is required to implement strain based dynamics for cloth simulation.//
//
//This code is under BSD license. If you make some improvements,
//or are using this in your research, do let me know and I would appreciate
//if you acknowledge this in your code or in your publication.
//
//Controls:
//Mouse: 
// - left click on any empty region to rotate, middle click to zoom 
// - left click and drag any point to drag it.
// - right click to pan camera
//
//Keyboard:
// - q/e keys to decrement/increment kSxx value  
// - a/d keys to decrement/increment kSyy value  
// - z/c keys to decrement/increment kSxy value   
// - p key to show/hide points
//
//Author: Dr. Muhammad Mobeen Movania
//        Assistant Professor 
//        Department of Computer Science
//        DHA Suffa University
//        Karachi, PAKISTAN
//
//Email: mova0002@e.ntu.edu.sg;mobeen.movania@dsu.edu.pk
//Last Modified: 20 September 2015

#include <GL/glew.h>
#include <GL/wglew.h>
#include <GL/freeglut.h>
#include <vector>
 
#define GLM_MESSAGES 
#define GLM_FORCE_PURE
#define GLM_FORCE_CXX98 
#include <glm/glm.hpp> 
#include <glm/gtc/matrix_transform.hpp> //for matrices
#include <glm/gtc/type_ptr.hpp> 

#include <cassert>
#include <iostream>

#include <minmax.h>
#define GL_CHECK_ERROR glGetError() == GL_NO_ERROR;

//undefine if u want to use the default bending constraint of strain based dynamics
#define USE_TRIANGLE_BENDING_CONSTRAINT_PBD

#pragma comment(lib, "glew32.lib")

using namespace std;  
const int width = 1024, height = 1024;

#define PI 3.1415926536f
#define EPSILON  0.0000001f

int numX = 20, numY=20; //these ar the number of quads
const size_t total_points = (numX+1)*(numY+1);
int size = 4;
float hsize = size/2.0f;

char info[MAX_PATH]={0};

float timeStep = 1.0f/60.0f;  
float currentTime = 0;
double accumulator = timeStep;
int selected_index = -1;
float global_dampening = 0.98f; //DevO: 24.07.2011  //global velocity dampening !!!

//distance constraint of position based dynamics
struct DistanceConstraint {	int p1, p2;	float rest_length, k; float k_prime; };

//triangle bending constraint of position based dynamics
struct TriangleBendingConstraint {	int p1, p2, p3;	float w,rest_length, k_prime;};

//bending constraint of strain based dynamics
struct BendingConstraint {	int p1, p2, p3, p4;	float k_prime;};

//triangle constraint of strain based dynamics 
struct TriangleConstraint {	int p1, p2, p3;	};

vector<GLushort> indices;

vector<DistanceConstraint> d_constraints;
  
#ifdef USE_TRIANGLE_BENDING_CONSTRAINT_PBD
vector<TriangleBendingConstraint> b_constraints;
#else
vector<BendingConstraint> b_constraints;
#endif

vector<float> phi0;			//initial dihedral angle between adjacent triangles

vector<TriangleConstraint> t_constraints;

//particle system
vector<glm::vec2> UV;		//texture coordinates
vector<glm::vec3> dp;		//positional corrections
 

vector<glm::mat2> Qinv;		//inverse of material coordinates
vector<glm::vec3> X;		//position (P in paper) 
vector<glm::vec3> tmp_X;	//predicted position
vector<glm::vec3> V;		//velocity
vector<glm::vec3> F;		//forces
vector<float> W;			//inverse particle mass 
vector<glm::vec3> Ri;		//Ri = Xi-Xcm 


DistanceConstraint mouse_constraint;
glm::vec3 mouse_position;

float tx, ty=-2, tz = -8;
int oldX=0, oldY=0;
float rX=15, rY=45;
int state =1 ; 
const int GRID_SIZE=10;

const size_t solver_iterations = 10; //number of solver iterations per step. PBD  

float kBend = 0.75f; 
float kStretch = 0.5f; 

float kSxx = 1.0f;
float kSyy = 1.0f;
float kSxy = 1.0f;
float kShear = 0.5f;
float kDamp = 0.000125f;
glm::vec3 gravity=glm::vec3(0.0f,-9.81f,0.0f);  

float mass = 1.f; 


GLint viewport[4];
GLdouble MV[16];
GLdouble P[16];

LARGE_INTEGER frequency;        // ticks per second
LARGE_INTEGER t1, t2;           // ticks
double frameTimeQP=0;
float frameTime =0 ;


glm::vec3 Up=glm::vec3(0,1,0), Right, viewDir;
float startTime =0, fps=0;
int totalFrames=0;

glm::mat4 ellipsoid, inverse_ellipsoid;
int iStacks = 30;
int iSlices = 30;
float fRadius = 1;
bool bDrawPoints = true;
// Resolve constraint in object space
glm::vec3 center = glm::vec3(0,0,0); //object space center of ellipsoid
float radius = 1;					 //object space radius of ellipsoid
float winZ=0; 

void StepPhysics(float dt);

float GetArea(int a, int b, int c) {
	glm::vec3 e1 = X[b]-X[a];
	glm::vec3 e2 = X[c]-X[a];
	return 0.5f * glm::length(glm::cross(e1,e2));
}
 
void AddDistanceConstraint(int a, int b, float k) {
	DistanceConstraint c;
	c.p1=a;
	c.p2=b;
	c.k =k;
	c.k_prime = 1.0f-pow((1.0f-c.k), 1.0f/solver_iterations);  //1.0f-pow((1.0f-c.k), 1.0f/ns);

	if(c.k_prime>1.0) 
		c.k_prime = 1.0;

	glm::vec3 deltaP = X[c.p1]-X[c.p2];
	c.rest_length = glm::length(deltaP);

	d_constraints.push_back(c);
} 

#ifdef USE_TRIANGLE_BENDING_CONSTRAINT_PBD
void AddBendingConstraint(int pa, int pb, int pc, float k) {
	TriangleBendingConstraint c;
	c.p1=pa;
	c.p2=pb;
	c.p3=pc; 
	
	c.w = W[pa] + W[pb] + 2*W[pc]; 
	glm::vec3 center = 0.3333f * (X[pa] + X[pb] + X[pc]);
 	c.rest_length = glm::length(X[pc]-center); 
	c.k_prime = 1.0f-pow((1.0f-k), 1.0f/solver_iterations);  
	if(c.k_prime>1.0) 
		c.k_prime = 1.0;
	b_constraints.push_back(c);
}
inline int getIndex(int i, int j) {
	return j*(numX+1) + i;
}

#else
void AddBendingConstraint(int pa, int pb, int pc,int pd, float k) {
	BendingConstraint c;
	c.p1=pa;
	c.p2=pb;
	c.p3=pc; 
	c.p4=pd;  
	
	c.k_prime = 1.0f-pow((1.0f-k), 1.0f/solver_iterations);  //1.0f-pow((1.0f-c.k), 1.0f/ns);
	if(c.k_prime>1.0) 
		c.k_prime = 1.0;
	b_constraints.push_back(c);
}
#endif

//this is the triangle constraint of strain based dynamics
void AddTriangleConstraint(int pa, int pb, int pc, float k) {
	TriangleConstraint c;
	c.p1=pa;
	c.p2=pb;
	c.p3=pc;   
	t_constraints.push_back(c);
}
void OnMouseDown(int button, int s, int x, int y)
{
	if (s == GLUT_DOWN ) 
	{
		oldX = x; 
		oldY = y;  

		if( button == GLUT_LEFT_BUTTON) {
			int window_y = (viewport[3] - y);   
			glReadPixels( x, window_y, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &winZ );


			double objX=0, objY=0, objZ=0;
			gluUnProject(x,window_y, winZ,  MV,  P, viewport, &objX, &objY, &objZ);
			glm::vec3 pt(objX,objY, objZ); 
			size_t i=0;
			for(i=0;i<total_points;i++) {			 
				if(glm::distance(X[i],pt)<0.1) {
					selected_index = i;
					printf("Intersected at %d\n",i);
					if(i!=0 && i!=numX )  {
						mouse_constraint.p1 = i; 
						mouse_constraint.p2 = -1;  

						mouse_position.x = (float)objX;
						mouse_position.y = (float)objY;
						mouse_position.z = (float)objZ;
					}
					break;
				}
			}
		}
	}	

	if(button == GLUT_MIDDLE_BUTTON)
		state = 0;
	else if(button == GLUT_RIGHT_BUTTON)
		state = 2;
	else
		state = 1;

	if(s==GLUT_UP) {
		selected_index= -1;
		mouse_constraint.p1 = -1;
		glutSetCursor(GLUT_CURSOR_INHERIT);
	}
}

void OnMouseMove(int x, int y)
{
	if(selected_index == -1) {
		if (state == 0)
			tz *= (1 + (y - oldY)/60.0f); 
		else if (state == 2) {
			tx -= (x - oldX)/50.0f; 
			ty += (y - oldY)/50.0f; 
		}
		else
		{
			rY += (x - oldX)/5.0f; 
			rX += (y - oldY)/5.0f; 
		} 
	} else {
		if(mouse_constraint.p1 ==-1) {
			float delta = 1500/abs(tz);
			float valX = (x - oldX)/delta; 
			float valY = (oldY - y)/delta; 
			if(abs(valX)>abs(valY))
				glutSetCursor(GLUT_CURSOR_LEFT_RIGHT);
			else 
				glutSetCursor(GLUT_CURSOR_UP_DOWN);


			
			V[selected_index] = glm::vec3(0.0f,0.0f,0.0f);
			X[selected_index].x += Right[0]*valX ;
			float newValue = X[selected_index].y+Up[1]*valY;
			if(newValue>0)
				X[selected_index].y = newValue;
			X[selected_index].z += Right[2]*valX + Up[2]*valY;	
		} else {
			double objX=0, objY=0, objZ=0; 
			int window_y = (viewport[3] - y); 
			gluUnProject(x, window_y, winZ,  MV,  P, viewport, &objX, &objY, &objZ);
			mouse_position.x = (float)objX;
			mouse_position.y = (float)objY;
			mouse_position.z = (float)objZ;
		}
	}
	oldX = x; 
	oldY = y; 

	glutPostRedisplay(); 
}


void DrawGrid()
{
	glBegin(GL_LINES);
	glColor3f(0.5f, 0.5f, 0.5f);
	for(int i=-GRID_SIZE;i<=GRID_SIZE;i++)
	{
		glVertex3f((float)i,0,(float)-GRID_SIZE);
		glVertex3f((float)i,0,(float)GRID_SIZE);

		glVertex3f((float)-GRID_SIZE,0,(float)i);
		glVertex3f((float)GRID_SIZE,0,(float)i);
	}
	glEnd();
}

inline glm::vec3 GetNormal(int ind0, int ind1, int ind2) {
	glm::vec3 e1 = X[ind0]-X[ind1];
	glm::vec3 e2 = X[ind2]-X[ind1];
	return glm::normalize(glm::cross(e1,e2));
}

inline float GetDihedralAngle(BendingConstraint c, float& d, glm::vec3& n1, glm::vec3& n2) {	 
	n1 = GetNormal(c.p1, c.p2, c.p3);
	n2 = GetNormal(c.p1, c.p2, c.p4); 
	d = glm::dot(n1, n2);
	return acos(d);
} 
void InitGL() { 

	mouse_constraint.p1 = -1;
	mouse_constraint.k = 0.5f;
	mouse_constraint.k_prime = 1.0f-pow((1.0f-mouse_constraint.k), 1.0f/solver_iterations);  //1.0f-pow((1.0f-c.k), 1.0f/ns);

	if(mouse_constraint.k_prime>1.0) 
		mouse_constraint.k_prime = 1.0;

	mouse_constraint.rest_length = 0.001f;


	startTime = (float)glutGet(GLUT_ELAPSED_TIME);
	currentTime = startTime;

	// get ticks per second
	QueryPerformanceFrequency(&frequency);

	// start timer
	QueryPerformanceCounter(&t1);


	glEnable(GL_DEPTH_TEST);
	size_t i=0, j=0, count=0;
	int l1=0, l2=0;
	float ypos = 7.0f;
	int v = numY+1;
	int u = numX+1;

	indices.resize( numX*numY*2*3);
	Qinv.resize(indices.size()/3); 
	X.resize(total_points);
	UV.resize(total_points);
	tmp_X.resize(total_points);
	dp.resize(total_points);
	V.resize(total_points);
	F.resize(total_points); 
	Ri.resize(total_points); 

	//fill in positions
	for(int j=0;j<=numY;j++) {		 
		for(int i=0;i<=numX;i++) {	 
			float tu = (float(i)/(u-1));
			float tv = (float(j)/(v-1));
			UV[count] = glm::vec2(tu,tv); 
			X[count] = glm::vec3( (tu*2-1)*hsize, size+1.0f, tv* size);
			count++;			
		}
	}

	///DevO: 24.07.2011
	W.resize(total_points); 
	std::fill(W.begin(), W.end(), 1.0f/mass);
	/// 2 Fixed Points 
	W[0] = 0.0;
	W[numX] = 0.0;
	 

	memcpy(&tmp_X[0].x, &X[0].x, sizeof(glm::vec3)*total_points);
	memcpy(&dp[0].x, &dp[0].x, sizeof(glm::vec3)*total_points);

	//fill in velocities	 
	memset(&(V[0].x),0,total_points*sizeof(glm::vec3));

	//fill in indices
	GLushort* id=&indices[0];
	for (int i = 0; i < numY; i++) {        
		for (int j = 0; j < numX; j++) {            
			int i0 = i * (numX+1) + j;            
			int i1 = i0 + 1;            
			int i2 = i0 + (numX+1);            
			int i3 = i2 + 1;            
			if ((j+i)%2) {                
				*id++ = i0; *id++ = i2; *id++ = i1;                
				*id++ = i1; *id++ = i2; *id++ = i3;            
			} else {                
				*id++ = i0; *id++ = i2; *id++ = i3;                
				*id++ = i0; *id++ = i3; *id++ = i1;            
			}        
		}    
	}

	count = 0;
	//calculate material coordinates using Eq. 37 page 9
	for(i=0;i<indices.size();i+=3) {
		int i0 = indices[i];
		int i1 = indices[i+1];
		int i2 = indices[i+2];
		glm::vec3 x0 = X[i0];
		glm::vec3 x1 = X[i1];
		glm::vec3 x2 = X[i2];

		glm::vec2 u0 = UV[i0];
		glm::vec2 u1 = UV[i1];
		glm::vec2 u2 = UV[i2];
		 
		glm::mat2x3 P = glm::mat2x3((x1-x0),(x2-x0)); 
		glm::mat2x2 U = glm::mat2x2((u1-u0),(u2-u0));  
		glm::mat2x2 Uinv = glm::inverse(U); 

		glm::mat2x3 T = P*Uinv;
		glm::vec3 n1(T[0]);
		glm::vec3 n2(T[1]);

		n1 = glm::normalize(n1);
		n2 = glm::normalize(n2);

		glm::mat2x2 C = glm::transpose(glm::mat2x3(n1,n2))*P;
		Qinv[count] = glm::inverse(C);   

		//for debugging Tu and Tv form a local orthonormal basis 
		//Tu should point in local X axis and should be output as red color
		//Tv should point in local Z axis and should be output as blue color 
		/*
		Tu[i0] += glm::vec3(T[0][0], T[0][1], T[0][2]);
		Tu[i1] += glm::vec3(T[0][0], T[0][1], T[0][2]);
		Tu[i2] += glm::vec3(T[0][0], T[0][1], T[0][2]);

		Tv[i0] += glm::vec3(T[1][0], T[1][1], T[1][2]);
		Tv[i1] += glm::vec3(T[1][0], T[1][1], T[1][2]);
		Tv[i2] += glm::vec3(T[1][0], T[1][1], T[1][2]);
		*/ 

		count++;
	}



	//glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	//glPolygonMode(GL_BACK, GL_LINE);
	glPointSize(5);

	wglSwapIntervalEXT(0);

	//check the damping values
	if(kStretch>1)		kStretch=1;
	if(kStretch<0)		kStretch=0;
	if(kBend>1)			kBend=1;
	if(kBend<0)			kBend=0;
	if(kDamp>1)			kDamp=1;
	if(kDamp<0)			kDamp=0;
	if(global_dampening>1)		global_dampening = 1;

	//setup distance constraints of pbd
	// Horizontal
	/*
	for (l1 = 0; l1 < v; l1++)	// v
	for (l2 = 0; l2 < (u - 1); l2++) {
	AddDistanceConstraint((l1 * u) + l2,(l1 * u) + l2 + 1, kStretch);
	}

	// Vertical
	for (l1 = 0; l1 < (u); l1++)	
	for (l2 = 0; l2 < (v - 1); l2++) {
	AddDistanceConstraint((l2 * u) + l1,((l2 + 1) * u) + l1, kStretch);
	}


	// Shearing distance constraint
	for (l1 = 0; l1 < (v - 1); l1++)	
	for (l2 = 0; l2 < (u - 1); l2++) {
	AddDistanceConstraint((l1 * u) + l2,((l1 + 1) * u) + l2 + 1, kStretch);
	AddDistanceConstraint(((l1 + 1) * u) + l2,(l1 * u) + l2 + 1, kStretch);
	}
	*/

	// create bending constraints	 
	#ifdef USE_TRIANGLE_BENDING_CONSTRAINT_PBD
		//add vertical constraints
		for(int i=0;i<=numX;i++) {
			for(int j=0;j<numY-1 ;j++) {
				AddBendingConstraint(getIndex(i,j), getIndex(i,(j+1)), getIndex(i,j+2), kBend);
			}
		}
		//add horizontal constraints
		for(int i=0;i<numX-1;i++) {
			for(int j=0;j<=numY;j++) {	   
				AddBendingConstraint(getIndex(i,j), getIndex(i+1,j), getIndex(i+2,j), kBend);
			}
		}
	#else
		//setup bending constraint of strain based dynamics
		for(int i = 0; i < v-1; ++i) {
			for(int j = 0; j < u-1; ++j) {	 			 
				int p1 = i * (numX+1) + j;            
				int p2 = p1 + 1;            
				int p3 = p1 + (numX+1);            
				int p4 = p3 + 1;   

				if ((j+i)%2) {  						
					AddBendingConstraint(p3,p2,p1,p4, kBend);	
				} else {  
					AddBendingConstraint(p4,p1,p3,p2, kBend);
				}     
			}
		}		 
		float d;
		glm::vec3 n1, n2;
		phi0.resize(b_constraints.size());

		for(i=0;i<b_constraints.size();i++) {		
			phi0[i] = GetDihedralAngle(b_constraints[i],d,n1,n2);		
		}	
	#endif

	//triangle constraints of strain based dynamics
	for(size_t i=0;i<indices.size();i+=3) {
		int i0 = indices[i];
		int i1 = indices[i+1];
		int i2 = indices[i+2];
		AddTriangleConstraint(i0,i1,i2, kStretch);
	}

	//create a basic ellipsoid object
	ellipsoid = glm::translate(glm::mat4(1.0f),glm::vec3(0.0f,2.0f,0.0f));
	ellipsoid = glm::rotate(ellipsoid, glm::radians(45.0f) ,glm::vec3(1.0f,0.0f,0.0f));
	ellipsoid = glm::scale(ellipsoid, glm::vec3(fRadius,fRadius,fRadius/2.0f));
	inverse_ellipsoid = glm::inverse(ellipsoid);

	//setup OpenGL lighting
	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHTING);
	glEnable(GL_DEPTH_TEST);

	GLfloat ambient[4]={0.25f,0.25f,0.25f,0.25f};
	GLfloat diffuse[4]={1,1,1,1}; 

	GLfloat mat_diffuse[4]={0.5,0.5,0.5,1};

	glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
	glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, mat_diffuse); 
	
	glDisable(GL_LIGHTING);
	glClearColor(1,1,1,1);

}

void OnReshape(int nw, int nh) {
	glViewport(0,0,nw, nh);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60, (GLfloat)nw / (GLfloat)nh, 1.f, 100.0f);

	glGetIntegerv(GL_VIEWPORT, viewport); 
	glGetDoublev(GL_PROJECTION_MATRIX, P);

	glMatrixMode(GL_MODELVIEW);
}

void OnRender() {		
	size_t i=0;
	float newTime = (float) glutGet(GLUT_ELAPSED_TIME);
	frameTime = newTime-currentTime;
	currentTime = newTime;
	//accumulator += frameTime;

	//Using high res. counter
	QueryPerformanceCounter(&t2);
	// compute and print the elapsed time in millisec
	frameTimeQP = (t2.QuadPart - t1.QuadPart) * 1000.0 / frequency.QuadPart;
	t1=t2;
	accumulator += frameTimeQP;

	++totalFrames;
	if((newTime-startTime)>1000)
	{		
		float elapsedTime = (newTime-startTime);
		fps = (totalFrames/ elapsedTime)*1000 ;
		startTime = newTime;
		totalFrames=0;
	}

	sprintf_s(info, "Click and drag any point to move it - FPS: %3.2f, Frame time (GLUT): %3.4f msecs, Frame time (QP): %3.3f", fps, frameTime, frameTimeQP);
	glutSetWindowTitle(info);

	glClear(GL_COLOR_BUFFER_BIT| GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();

	//set viewing transformation
	glTranslatef(tx,ty,tz); 
	glRotatef(rX,1,0,0);
	glRotatef(rY,0,1,0);

	glGetDoublev(GL_MODELVIEW_MATRIX, MV);
	viewDir.x = (float)-MV[2];
	viewDir.y = (float)-MV[6];
	viewDir.z = (float)-MV[10];
	Right = glm::cross(viewDir, Up);

	//draw grid
	DrawGrid();

	//draw ellipsoid
	glColor3f(0,1,0);
	glPushMatrix();
		glMultMatrixf(glm::value_ptr(ellipsoid));
		glutWireSphere(fRadius, iSlices, iStacks);
	glPopMatrix();
	

	//draw polygons
	glColor3f(1,1,1);
	glEnable(GL_LIGHTING);
	
	glBegin(GL_TRIANGLES);
	int count = 0;
	for(i=0;i<indices.size();i+=3) {
		int i0 = indices[i];
		int i1 = indices[i+1];
		int i2 = indices[i+2];

		glm::vec3 p1 = X[i0];
		glm::vec3 p2 = X[i1];
		glm::vec3 p3 = X[i2];

		glm::vec2 uv1 = UV[i0];
		glm::vec2 uv2 = UV[i1];
		glm::vec2 uv3 = UV[i2];

		glm::vec3 N = glm::normalize(glm::cross((p2-p1), (p3-p1)));
		//material coordinates
		//glm::mat2 q1 = Qinv[i0];
		//glm::mat2 q2 = Qinv[i1];
		//glm::mat2 q3 = Qinv[i2];

		//for debugging
		/*
		glm::vec3 tu1 = Tu[i0];
		glm::vec3 tu2 = Tu[i1];
		glm::vec3 tu3 = Tu[i2];

		glm::vec3 tv1 = Tv[i0];
		glm::vec3 tv2 = Tv[i1];
		glm::vec3 tv3 = Tv[i2];
		*/


		//output strains

		//glm::mat2x2 St = S[count++]; 

		//glColor3f(St[0][0], St[1][1], St[1][0]);	
		glNormal3f(N.x, N.y, N.z);
		glVertex3f(p1.x,p1.y,p1.z);		
		glVertex3f(p2.x,p2.y,p2.z);		
		glVertex3f(p3.x,p3.y,p3.z);
	}
	glEnd();	 
	glDisable(GL_LIGHTING);


	if(bDrawPoints) {
		//draw points	
		glBegin(GL_POINTS);
		for(i=0;i<total_points;i++) {
			glm::vec3 p = X[i];
			int is = (i==selected_index);
			glColor3f((float)!is,(float)is,(float)is);
			glVertex3f(p.x,p.y,p.z);
		} 
		glEnd();
	}


	//draw mouse line
	if(mouse_constraint.p1 != -1) {
		glBegin(GL_LINES);
		glColor3f(0,1,1);
		glVertex3fv(&X[mouse_constraint.p1].x);
		glVertex3f(mouse_position.x, mouse_position.y, mouse_position.z);
		glEnd();
	}

	//draw normals for debug only 	
#ifndef USE_TRIANGLE_BENDING_CONSTRAINT_PBD
#ifdef _DEBUG
	BendingConstraint b;
	float size = 0.1f;
	float d = 0;
	glm::vec3 n1, n2, c1, c2;


	glBegin(GL_LINES);
	for(i=0;i<b_constraints.size();i++) {
		b = b_constraints[i];
		c1 = (X[b.p1] + X[b.p2] + X[b.p3]) /3.0f;
		c2 = (X[b.p1] + X[b.p2] + X[b.p4]) /3.0f;
		GetDihedralAngle(b,d,n1,n2);
		glColor3f(abs(n1.x), abs(n1.y), abs(n1.z) );
		glVertex3f(c1.x,c1.y,c1.z);		glVertex3f(c1.x+size*n1.x,c1.y+size*n1.y,c1.z+size*n1.z);

		glColor3f(abs(n2.x), abs(n2.y), abs(n2.z));
		glVertex3f(c2.x,c2.y,c2.z);		glVertex3f(c2.x+size*n2.x,c2.y+size*n2.y,c2.z+size*n2.z);
	}
	glEnd();
#endif
#endif
	glutSwapBuffers();
}

void OnShutdown() {	
//	d_constraints.clear();
	b_constraints.clear();
	t_constraints.clear();
	indices.clear();
	X.clear();
	F.clear();
	V.clear();
	phi0.clear();
	W.clear();
	tmp_X.clear();
	dp.clear();
	Ri.clear();
}

void ComputeForces( ) {
	size_t i=0;

	for(i=0;i<total_points;i++) {
		F[i] = glm::vec3(0); 

		//add gravity force
		if(W[i]>0)		 
			F[i] += gravity/W[i] ; 
	}	
} 
void IntegrateExplicitWithDamping(float deltaTime) {
	float deltaTimeMass = deltaTime;
	size_t i=0;

	glm::vec3 Xcm = glm::vec3(0);
	glm::vec3 Vcm = glm::vec3(0);
	float sumM = 0;
	for(i=0;i<total_points;i++) {

		V[i] *= global_dampening; //global velocity dampening !!!		
		V[i] = V[i] + (F[i]*deltaTime)*W[i]; 	 					

		//calculate the center of mass's position 
		//and velocity for damping calc
		Xcm += (X[i]*mass);
		Vcm += (V[i]*mass);
		sumM += mass;
	}
	Xcm /= sumM; 
	Vcm /= sumM; 

	glm::mat3 I = glm::mat3(1.0f);
	glm::vec3 L = glm::vec3(0.0f,0.0f,0.0f);
	glm::vec3 w = glm::vec3(0.0f,0.0f,0.0f);//angular velocity


	for(i=0;i<total_points;i++) { 
		Ri[i] = (X[i] - Xcm);	

		L += glm::cross(Ri[i],mass*V[i]); 

		//thanks to DevO for pointing this and these notes really helped.
		//http://www.sccg.sk/~onderik/phd/ca2010/ca10_lesson11.pdf

		glm::mat3 tmp = glm::mat3(0,-Ri[i].z,  Ri[i].y, 
			Ri[i].z,       0,-Ri[i].x,
			-Ri[i].y,Ri[i].x,        0);
		I +=(tmp*glm::transpose(tmp))*mass;
	} 

	w = glm::inverse(I)*L;

	//apply center of mass damping
	for(i=0;i<total_points;i++) {
		glm::vec3 delVi = Vcm + glm::cross(w,Ri[i])-V[i];		
		V[i] += kDamp*delVi;
	}

	//calculate predicted position
	for(i=0;i<total_points;i++) {
		if(W[i] <= 0.0) { 
			tmp_X[i] = X[i]; //fixed points
		} else {
			tmp_X[i] = X[i] + (V[i]*deltaTime);				 
		}
	} 
} 
void IntegrateExplicitWithDampingSBD(float deltaTime) {
	size_t i=0; 
	float sumVN=0;

	for(i=0;i<total_points;i++) {
		V[i] *= global_dampening; //global velocity dampening !!!		
		V[i] = V[i] + (F[i]*deltaTime)*W[i];
		float lenDP = glm::length(dp[i]);
		if(lenDP > EPSILON)
			sumVN = sumVN + glm::dot(V[i], dp[i]/lenDP);
	}

	for(i=0;i<total_points;i++) {
		float lenDP = glm::length(dp[i]);
		if(lenDP > EPSILON)
			V[i] = V[i] - kDamp*sumVN* dp[i]/lenDP;
	}

	//calculate predicted position
	for(i=0;i<total_points;i++) {
		if(W[i] <= 0.0) { 
			tmp_X[i] = X[i]; //fixed points
		} else {
			tmp_X[i] = X[i] + (V[i]*deltaTime);				 
		}
	} 
}
void Integrate(float deltaTime) {	
	float inv_dt = 1.0f/deltaTime;
	size_t i=0; 

	for(i=0;i<total_points;i++) {	
		V[i] = (tmp_X[i] - X[i])*inv_dt;		
		X[i] = tmp_X[i];		 
	}
}
/*
void UpdateDistanceConstraint(int i) {

	DistanceConstraint c = d_constraints[i];
	glm::vec3 dir = tmp_X[c.p1] - tmp_X[c.p2];

	float len = glm::length(dir); 
	if(len <= EPSILON) 
		return;

	float w1 = W[c.p1];
	float w2 = W[c.p2];
	float invMass = w1+ w2; 
	if(invMass <= EPSILON) 
		return;

	glm::vec3 dP = (1.0f/invMass) * (len-c.rest_length ) * (dir/len)* c.k_prime;
	if(w1 > 0.0)
		tmp_X[c.p1] -= dP*w1;

	if(w2 > 0.0)
		tmp_X[c.p2] += dP*w2;	
}
*/
void UpdateBendingConstraint(int index) {
	size_t i=0;
	#ifdef USE_TRIANGLE_BENDING_CONSTRAINT_PBD
		//Main reference
		//http://image.diku.dk/kenny/download/kelager.niebe.ea10.pdf
 
		TriangleBendingConstraint c = b_constraints[index];

		//global_k is a percentage of the global dampening constant 
		float global_k = global_dampening*0.01f; 
		glm::vec3 center = 0.3333f * (tmp_X[c.p1] + tmp_X[c.p2] + tmp_X[c.p3]);
		glm::vec3 dir_center = tmp_X[c.p3]-center;
		float dist_center = glm::length(dir_center);

		float diff = 1.0f - ((global_k + c.rest_length) / dist_center);
		glm::vec3 dir_force = dir_center * diff;
		glm::vec3 fa = c.k_prime * ((2.0f*W[c.p1])/c.w) * dir_force;
		glm::vec3 fb = c.k_prime * ((2.0f*W[c.p2])/c.w) * dir_force;
		glm::vec3 fc = -c.k_prime * ((4.0f*W[c.p3])/c.w) * dir_force;

		if(W[c.p1] > 0.0)  {
			tmp_X[c.p1] += fa;
		}
		if(W[c.p2] > 0.0) {
			tmp_X[c.p2] += fb;
		}
		if(W[c.p3] > 0.0) {
			tmp_X[c.p3] += fc;
		}

	#else

		BendingConstraint c = b_constraints[index];  

		//Using the dihedral angle approach of the position based dynamics		
		float d = 0, phi=0,i_d=0;

		glm::vec3 p1 = tmp_X[c.p1];
		glm::vec3 p2 = tmp_X[c.p2]-p1;
		glm::vec3 p3 = tmp_X[c.p3]-p1;
		glm::vec3 p4 = tmp_X[c.p4]-p1;

		glm::vec3 p3p1 = p3-p1;
		glm::vec3 p4p1 = p4-p1;
		glm::vec3 p3p2 = p3-p2;
		glm::vec3 p4p2 = p4-p2;

		//prepare e, n1 and n2 using Eqs. 44, 45, 46
		glm::vec3 e  = p4-p3;		
		glm::vec3 n1 = glm::cross(p3p1,p4p1);
		float  denN1 = glm::dot(n1, n1);		//glm::length(n1)*glm::length(n1);

		if(denN1 <= EPSILON) { return; } //need to handle this case.

		n1 = n1/denN1;

		glm::vec3 n2 = glm::cross(p4p2,p3p2);
		float  denN2 = glm::dot(n2,n2);
		if(denN2 <= EPSILON) { return; } //need to handle this case.

		n2 = n2/denN2;

		d	= glm::dot(n1,n2);
		phi = acos(d);

		float sign = (glm::dot(glm::cross(n1,n2),e) > 0.0f)?-1.0f:1.0f;

		//try to catch invalid values that will return NaN.
		// sqrt(1 - (1.0001*1.0001)) = NaN 
		// sqrt(1 - (-1.0001*-1.0001)) = NaN 
		if(d<-1.0) 
			d = -1.0; 
		else if(d>1.0) 
			d=1.0; //d = clamp(d,-1.0,1.0);

		//in both case sqrt(1-d*d) will be zero and nothing will be done.
		//0 case, the triangles are facing in the opposite direction, folded together.
		if(d == -1.0){ 
			phi = PI;  //acos(-1.0) == PI
			if(phi == phi0[index]) 
				return; //nothing to do 

			//in this case one just need to push 
			//vertices 1 and 2 in n1 and n2 directions, 
			//so the constrain will do the work in second iterations.
			//if(c.p1!=0 && c.p1!=numX)
			//	tmp_X[c.p1] += sign*n1/1000.0f;

			//if(c.p2!=0 && c.p2!=numX)
			//	tmp_X[c.p4] += sign*n2/1000.0f;

			return;
		}
		if(d == 1.0){ //180 case, the triangles are planar
			phi = 0.0;  //acos(1.0) == 0.0
			if(phi == phi0[index]) 
				return; //nothing to do 
		}


		i_d = sqrt(1-(d*d))*(phi-phi0[index]) ;

		float lenE = glm::length(e);
		if(lenE <= EPSILON)
			return;

		glm::vec3 p1p4 = -p4p1;
		glm::vec3 p2p4 = -p4p2; 

		glm::vec3 q1 = lenE*n1*sign;	//Eq. 40
		glm::vec3 q2 = lenE*n2*sign;	//Eq. 41
		glm::vec3 q3 = ((glm::dot(p1p4, e)/lenE)*n1 + (glm::dot(p2p4, e)/lenE)*n2)*sign; //Eq. 42
		glm::vec3 q4 = ((glm::dot(p3p1, e)/lenE)*n1 + (glm::dot(p3p2, e)/lenE)*n2)*sign; //Eq. 43

		float q1_len2 = glm::dot(q1,q1); 
		float q2_len2 = glm::dot(q2,q2); 
		float q3_len2 = glm::dot(q3,q3); 
		float q4_len2 = glm::dot(q4,q4); 

		float sum = W[c.p1]*(q1_len2) + W[c.p2]*(q2_len2) +	W[c.p3]*(q3_len2) +	W[c.p4]*(q4_len2);	
		if(sum <= EPSILON)
			return;
		glm::vec3 dP1 = -( (W[c.p1] * i_d) /sum)*q1;
		glm::vec3 dP2 = -( (W[c.p2] * i_d) /sum)*q2;
		glm::vec3 dP3 = -( (W[c.p3] * i_d) /sum)*q3;
		glm::vec3 dP4 = -( (W[c.p4] * i_d) /sum)*q4;

		if(W[c.p1] > 0.0) {
			tmp_X[c.p1] += dP1*c.k_prime;
		}
		if(W[c.p2] > 0.0) {
			tmp_X[c.p2] += dP2*c.k_prime;
		}
		if(W[c.p3] > 0.0) {
			tmp_X[c.p3] += dP3*c.k_prime;
		}	
		if(W[c.p4] > 0.0) {
			tmp_X[c.p4] += dP4*c.k_prime;
		}
	#endif
}
 
glm::mat3x2 outerProduct(const glm::vec3& a, const glm::vec2& b) {
	glm::mat3x2 ret;

	ret[0][0] = a.x * b.x;
	ret[1][0] = a.y * b.x;
	ret[2][0] = a.z * b.x;
	ret[0][1] = a.x * b.y;
	ret[1][1] = a.y * b.y;
	ret[2][1] = a.z * b.y;

	return ret;
}

void UpdateTriangleConstraint(int index) {
	TriangleConstraint c = t_constraints[index];
	float w1 = W[c.p1];
	float w2 = W[c.p2];
	float w3 = W[c.p3];   

	glm::mat2x2 C   = Qinv[index];	 

	glm::vec3 x0 = tmp_X[c.p1];
	glm::vec3 x1 = tmp_X[c.p2];
	glm::vec3 x2 = tmp_X[c.p3];

	glm::vec2 C1 = C[0]; 
	glm::vec2 C2 = C[1]; 

	glm::mat2x3 P =(glm::mat2x3((x1-x0),(x2-x0))); 
		
	glm::mat2x3 F = P*C;

	glm::vec3 f1(F[0]); 
	glm::vec3 f2(F[1]); 

	//Eq. 28 page 6
	float S11 = glm::dot(f1, f1);
	float S12 = glm::dot(f1, f2); 
	float S22 = glm::dot(f2, f2);

	//Eq. 29 page 6
	glm::mat3x2 mat11 = (2.0f*outerProduct(f1, C1));
	glm::mat3x2 mat12 = (outerProduct(f2, C1) + outerProduct(f1, C2));  
	glm::mat3x2 mat22 = (2.0f*outerProduct(f2, C2));

	//stretch constraints
	glm::vec3 ds1_dp1(mat11[0][0], mat11[1][0], mat11[2][0]); 
	glm::vec3 ds1_dp2(mat11[0][1], mat11[1][1], mat11[2][1]); 
	glm::vec3 ds1_dp0 = -(ds1_dp1+ds1_dp2);

	glm::vec3 ds2_dp1(mat22[0][0], mat22[1][0], mat22[2][0]);
	glm::vec3 ds2_dp2(mat22[0][1], mat22[1][1], mat22[2][1]); 
	glm::vec3 ds2_dp0 = -(ds2_dp1+ds2_dp2); 

	//shear constraints
	glm::vec3 ds12_dp1(mat12[0][0], mat12[1][0], mat12[2][0]); 
	glm::vec3 ds12_dp2(mat12[0][1], mat12[1][1], mat12[2][1]);  
	glm::vec3 ds12_dp0 = -(ds12_dp1+ds12_dp2); 

	glm::vec3 dp0(0,0,0);
	glm::vec3 dp1(0,0,0);
	glm::vec3 dp2(0,0,0);

	// 1) stretch - S11 See Eq. 33 page 7
	float weightSum11 = w1*glm::dot(ds1_dp0, ds1_dp0) + 
						w2*glm::dot(ds1_dp1, ds1_dp1) + 
						w3*glm::dot(ds1_dp2, ds1_dp2) ;  

	float lambda11 = (S11 - 1)/weightSum11;
	S11 = -kSxx * lambda11;

	dp0 += w1*S11*ds1_dp0; 
	dp1 += w2*S11*ds1_dp1; 
	dp2 += w3*S11*ds1_dp2; 

	// = 2) stretch - S22 See Eq. 33 page 7
	float weightSum22 = w1 * glm::dot(ds2_dp0, ds2_dp0) + 
						w2 * glm::dot(ds2_dp1, ds2_dp1) + 
						w3 * glm::dot(ds2_dp2, ds2_dp2)  ;

	float lambda22 = (S22 - 1)/weightSum22;
	S22 = -kSyy * lambda22; 
	dp0 += w1*S22*ds2_dp0; 
	dp1 += w2*S22*ds2_dp1; 
	dp2 += w3*S22*ds2_dp2;  

	// = 4) Shear - S12 See Eq. 34 page 7
	float weightSum12 = w1 * glm::dot(ds12_dp0, ds12_dp0) + 
						w2 * glm::dot(ds12_dp1, ds12_dp1) +
						w3 * glm::dot(ds12_dp2, ds12_dp2) ;
	float lambda12 = S12/weightSum12;
	S12 = -kSxy * lambda12; 

	dp0 += w1*S12*ds12_dp0; 
	dp1 += w2*S12*ds12_dp1; 
	dp2 += w3*S12*ds12_dp2; 

	if(w1 > EPSILON) {
		dp[c.p1] += dp0; 
		tmp_X[c.p1] += dp0;
	}

	if(w2 > EPSILON) {
		dp[c.p2] += dp1;
		tmp_X[c.p2] += dp1;
	}

	if(w3 > EPSILON) {
		dp[c.p3] += dp2;
		tmp_X[c.p3] += dp2;			 
	}	
}

//----------------------------------------------------------------------------------------------------
void GroundCollision() 
{
	for(size_t i=0;i<total_points;i++) {	
		if(tmp_X[i].y<0) //collision with ground
			tmp_X[i].y=0;
	}
}
void EllipsoidCollision() {
	for(size_t i=0;i<total_points;i++) {
		glm::vec4 X_0 = (inverse_ellipsoid*glm::vec4(tmp_X[i],1));
		glm::vec3 delta0 = glm::vec3(X_0.x, X_0.y, X_0.z) - center;
		float distance = glm::length(delta0);
		if (distance < 1.0f) {
			delta0 = (radius - distance) * delta0 / distance; 
			glm::vec3 delta = glm::vec3(ellipsoid * glm::vec4(delta0, 0));

			tmp_X[i] +=  delta; 

			V[i] = glm::vec3(0);
		} 
	}
}
void UpdateExternalConstraints() {
	EllipsoidCollision();
	//handle ground collision
	GroundCollision();
}
//----------------------------------------------------------------------------------------------------
void UpdateInternalConstraints() {
	size_t i=0;
	memset(&dp[0].x, 0, sizeof(glm::vec3)*total_points);

	//printf(" UpdateInternalConstraints \n ");
	for (size_t si=0;si<solver_iterations;++si) {
		 

		//update all triangle constraints
		for(i=0;i<t_constraints.size();++i) {
			UpdateTriangleConstraint(i);
		}

		//update all bending constraints
		for(i=0;i<b_constraints.size();i++) {
			UpdateBendingConstraint(i);
		} 

		//update mouse spring 
		if(mouse_constraint.p1 !=-1) {
			DistanceConstraint c = mouse_constraint;

			glm::vec3 dir = tmp_X[c.p1] - mouse_position;
			float len = glm::length(dir); 
			if(len <= EPSILON) 
				return;
			float w1 = W[c.p1]; 
			 
			glm::vec3 dP = (1.0f/w1) * (len-c.rest_length ) * (dir/len)* c.k_prime;
			tmp_X[c.p1] -= dP;
			 
		} 
	}
}

void OnIdle() {	

	/*
	//Semi-fixed time stepping
	if ( frameTime > 0.0 )
	{
	const float deltaTime = min( frameTime, timeStep );
	StepPhysics(deltaTime );
	frameTime -= deltaTime;    		
	}
	*/

	//printf(" ### OnIdle %f ### \n",accumulator);
	//Fixed time stepping + rendering at different fps	
	if ( accumulator >= timeStep )
	{	 
		StepPhysics(timeStep );		
		accumulator -= timeStep;
	}

	glutPostRedisplay(); 
	Sleep(5); //TODO
}

void StepPhysics(float dt ) {
	ComputeForces();	
	//IntegrateExplicitWithDamping(dt);
	IntegrateExplicitWithDampingSBD(dt); //for strain based dynamics 
	   
	UpdateInternalConstraints();	
	UpdateExternalConstraints();

	Integrate(dt);
}

void OnKey(unsigned char key, int, int) {
	switch(key) {
		case 'a': kSyy -= 0.1f; break;
		case 'd': kSyy += 0.1f; break;

		case 'q': kSxx -= 0.1f; break;
		case 'e': kSxx += 0.1f; break;

		case 'z': kSxy -= 0.1f; break;
		case 'c': kSxy += 0.1f; break;
		case 'p': bDrawPoints=!bDrawPoints; break; 
	}
	kSxx = max(0.1f, min(kSxx, 1.0f));
	kSxy = max(0.1f, min(kSxy, 1.0f));
	kSyy = max(0.1f, min(kSyy, 1.0f));

	printf("kSxx: %3.2f, kSyy: %3.2f, kSxy %3.2f\n", kSxx, kSyy, kSxy);
	glutPostRedisplay();
}

void main(int argc, char** argv) {

	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
	glutInitWindowSize(width, height);
	glutCreateWindow("GLUT Cloth Demo [Strain based Dynamics]");

	glutDisplayFunc(OnRender);
	glutReshapeFunc(OnReshape);
	glutIdleFunc(OnIdle);

	glutMouseFunc(OnMouseDown);
	glutMotionFunc(OnMouseMove);	
	glutKeyboardFunc(OnKey);
	glutCloseFunc(OnShutdown);

	glewInit();
	InitGL();

	glutMainLoop();		
}

